<?php

namespace FoodSearch;

/**
 * User_Geolocation Class.
 */
class User_Geolocation {

    /**
     * API endpoints for looking up user IP address.
     *
     * @var array
     */
    private static $ip_lookup_apis = array(
        'ipify' => 'http://api.ipify.org/',
        'ipecho' => 'http://ipecho.net/plain',
        'ident' => 'http://ident.me',
        'whatismyipaddress' => 'http://bot.whatismyipaddress.com',
    );

    /**
     * API endpoints for geolocating an IP address
     *
     * @var array
     */
    private static $geoip_apis = array(
        'ipinfo.io' => 'https://ipinfo.io/%s/json',
        'ip-api.com' => 'http://ip-api.com/json/%s',
    );

    /**
     * Get current user IP Address.
     *
     * @return string
     */
    public static function get_ip_address() {
        if (isset($_SERVER["HTTP_CF_CONNECTING_IP"])) {
            return sanitize_text_field(wp_unslash($_SERVER["HTTP_CF_CONNECTING_IP"]));
        } elseif (isset($_SERVER['HTTP_X_REAL_IP'])) {
            return sanitize_text_field(wp_unslash($_SERVER['HTTP_X_REAL_IP']));
        } elseif (isset($_SERVER['HTTP_X_FORWARDED_FOR'])) {
            // Make sure we always only send through the first IP in the list which should always be the client IP.
            return (string) rest_is_ip_address(trim(current(preg_split('/,/', sanitize_text_field(wp_unslash($_SERVER['HTTP_X_FORWARDED_FOR']))))));
        } elseif (isset($_SERVER['REMOTE_ADDR'])) {
            return sanitize_text_field(wp_unslash($_SERVER['REMOTE_ADDR']));
        }
        return '';
    }

    /**
     * Get user IP Address using an external service.
     * This can be used as a fallback for users on localhost where
     * get_ip_address() will be a local IP and non-geolocatable.
     *
     * @return string
     */
    public static function get_external_ip_address() {
        $external_ip_address = '0.0.0.0';

        if ('' !== self::get_ip_address()) {
            $transient_name = 'foodsearch_ext_ip_address_' . self::get_ip_address();
            $external_ip_address = get_transient($transient_name);
        }

        if (false === $external_ip_address) {
            $external_ip_address = '0.0.0.0';
            $ip_lookup_services = apply_filters('foodsearch_geolocation_ip_lookup_apis', self::$ip_lookup_apis);
            $ip_lookup_services_keys = array_keys($ip_lookup_services);
            shuffle($ip_lookup_services_keys);

            foreach ($ip_lookup_services_keys as $service_name) {
                $service_endpoint = $ip_lookup_services[$service_name];
                $response = wp_safe_remote_get($service_endpoint, array('timeout' => 2));

                if (!is_wp_error($response) && rest_is_ip_address($response['body'])) {
                    $external_ip_address = apply_filters('foodsearch_geolocation_ip_lookup_api_response', $response['body'], $service_name);
                    break;
                }
            }

            set_transient($transient_name, $external_ip_address, DAY_IN_SECONDS);
        }

        return $external_ip_address;
    }

    /**
     * Geolocate an IP address.
     *
     * @param  string $ip_address   IP Address.
     * @param  bool   $fallback     If true, fallbacks to alternative IP detection (can be slower).
     * @param  bool   $api_fallback If true, uses geolocation APIs if the database file doesn't exist (can be slower).
     * @return array
     */
    public static function geolocate_ip($ip_address = '', $fallback = false, $api_fallback = true) {
        
        if (empty($ip_address)) {
            $ip_address = self::get_ip_address();
        }

        /**
         * Get geolocation filter.
         *
         * @param array  $geolocation Geolocation data, including country, state, city, and postcode.
         * @param string $ip_address  IP Address.
         */
        $geolocation = apply_filters(
                'foodsearch_get_user_geolocation', array(
            'country' => '',
            'city' => '',
            'lat' => '',
            'lng' => '',
                ), $ip_address
        );

        // If we still haven't found a country code, let's consider doing an API lookup.
        if ('' === $geolocation['country'] && $api_fallback) {
            $geolocation = self::geolocate_via_api($ip_address);
        }

        // It's possible that we're in a local environment, in which case the geolocation needs to be done from the
        // external address.
        if ('' === $geolocation['country'] && $fallback) {
            $external_ip_address = self::get_external_ip_address();

            // Only bother with this if the external IP differs.
            if ('0.0.0.0' !== $external_ip_address && $external_ip_address !== $ip_address) {
                return self::geolocate_ip($external_ip_address, false, $api_fallback);
            }
        }

        return array(
            'country' => $geolocation['country'],
            'city' => $geolocation['city'],
            'lat' => $geolocation['lat'],
            'lng' => $geolocation['lng'],
        );
    }

    /**
     * Use APIs to Geolocate the user.
     *
     * Geolocation APIs can be added through the use of the foodsearch_geolocation_geoip_apis filter.
     * Provide a name=>value pair for service-slug=>endpoint.
     *
     * If APIs are defined, one will be chosen at random to fulfil the request. After completing, the result
     * will be cached in a transient.
     *
     * @param  string $ip_address IP address.
     * @return string
     */
    private static function geolocate_via_api($ip_address) {
        $loc_trans_data = get_transient('foodsearch_geo_loc_' . $ip_address);

        $loc_data = array(
            'country' => '',
            'city' => '',
            'lat' => '',
            'lng' => '',
        );
        if (false === $loc_trans_data) {
            $geoip_services = apply_filters('foodsearch_geolocation_geoip_apis', self::$geoip_apis);

            if (empty($geoip_services)) {
                return '';
            }

            $geoip_services_keys = array_keys($geoip_services);

            shuffle($geoip_services_keys);

            foreach ($geoip_services_keys as $service_name) {
                $service_endpoint = $geoip_services[$service_name];
                $response = wp_safe_remote_get(sprintf($service_endpoint, $ip_address), array('timeout' => 30));

                if (!is_wp_error($response) && $response['body']) {
                    switch ($service_name) {
                        case 'ipinfo.io':
                            $data = json_decode($response['body'], true);
                            $ip_loc = isset($data['loc']) ? $data['loc'] : '';
                            $ip_loc = $ip_loc != '' ? explode(',', $ip_loc) : '';
                            $loc_data = array(
                                'country' => isset($data['country']) ? $data['country'] : '',
                                'city' => isset($data['city']) ? $data['city'] : '',
                                'lat' => isset($ip_loc[0]) ? $ip_loc[0] : '',
                                'lng' => isset($ip_loc[1]) ? $ip_loc[1] : '',
                            );
                            break;
                        case 'ip-api.com':
                            $data = json_decode($response['body'], true);
                            $loc_data = array(
                                'country' => isset($data['country']) ? $data['country'] : '',
                                'city' => isset($data['city']) ? $data['city'] : '',
                                'lat' => isset($data['lat']) ? $data['lat'] : '',
                                'lng' => isset($data['lon']) ? $data['lon'] : '',
                            );
                            break;
                        default:
                            $data = array();
                            break;
                    }

                    if ($loc_data['lat'] != '' && $loc_data['lng'] != '') {
                        set_transient('foodsearch_geo_loc_' . $ip_address, $loc_data, DAY_IN_SECONDS);
                        break;
                    }
                }
            }
        } else {
            $loc_data = $loc_trans_data;
        }

        return $loc_data;
    }

}
